"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const cli_1 = require("@yarnpkg/cli");
const core_1 = require("@yarnpkg/core");
const core_2 = require("@yarnpkg/core");
const core_3 = require("@yarnpkg/core");
const clipanion_1 = require("clipanion");
const get_1 = __importDefault(require("lodash/get"));
const set_1 = __importDefault(require("lodash/set"));
const unset_1 = __importDefault(require("lodash/unset"));
const Constraints_1 = require("../Constraints");
// eslint-disable-next-line arca/no-default-export
class ConstraintsCheckCommand extends cli_1.BaseCommand {
    constructor() {
        super(...arguments);
        this.fix = false;
    }
    async execute() {
        for (let t = 0; t < 10; ++t) {
            const configuration = await core_1.Configuration.find(this.context.cwd, this.context.plugins);
            const { project } = await core_1.Project.find(configuration, this.context.cwd);
            const constraints = await Constraints_1.Constraints.find(project);
            const toSave = new Set();
            const report = await core_2.StreamReport.start({
                configuration,
                stdout: this.context.stdout,
            }, async (report) => {
                const result = await constraints.process();
                await processDependencyConstraints(toSave, result.enforcedDependencies, {
                    fix: this.fix,
                    configuration,
                    report,
                });
                await processFieldConstraints(toSave, result.enforcedFields, {
                    fix: this.fix,
                    configuration,
                    report,
                });
                for (const workspace of toSave) {
                    workspace.persistManifest();
                }
            });
            if (report.hasErrors())
                return report.exitCode();
            if (!this.fix || toSave.size > 0) {
                return 0;
            }
        }
        return 1;
    }
}
ConstraintsCheckCommand.usage = clipanion_1.Command.Usage({
    category: `Constraints-related commands`,
    description: `check that the project constraints are met`,
    details: `
      This command will run constraints on your project and emit errors for each one that is found but isn't met. If any error is emitted the process will exit with a non-zero exit code.

      If the \`--fix\` flag is used, Yarn will attempt to automatically fix the issues the best it can, following a multi-pass process (with a maximum of 10 iterations). Some ambiguous patterns cannot be autofixed, in which case you'll have to manually specify the right resolution.

      For more information as to how to write constraints, please consult our dedicated page on our website: https://yarnpkg.com/features/constraints.
    `,
    examples: [[
            `Check that all constraints are satisfied`,
            `yarn constraints`,
        ], [
            `Autofix all unmet constraints`,
            `yarn constraints --fix`,
        ]],
});
__decorate([
    clipanion_1.Command.Boolean(`--fix`)
], ConstraintsCheckCommand.prototype, "fix", void 0);
__decorate([
    clipanion_1.Command.Path(`constraints`)
], ConstraintsCheckCommand.prototype, "execute", null);
exports.default = ConstraintsCheckCommand;
async function processDependencyConstraints(toSave, enforcedDependencies, { configuration, fix, report }) {
    let hasFixes = false;
    const allIdents = new Map();
    const byWorkspaces = new Map();
    for (const { workspace, dependencyIdent, dependencyRange, dependencyType } of enforcedDependencies) {
        let byWorkspacesStore = byWorkspaces.get(workspace);
        if (typeof byWorkspacesStore === `undefined`)
            byWorkspaces.set(workspace, byWorkspacesStore = new Map());
        let byIdentStore = byWorkspacesStore.get(dependencyIdent.identHash);
        if (typeof byIdentStore === `undefined`)
            byWorkspacesStore.set(dependencyIdent.identHash, byIdentStore = new Map());
        let byDependencyTypeStore = byIdentStore.get(dependencyType);
        if (typeof byDependencyTypeStore === `undefined`)
            byIdentStore.set(dependencyType, byDependencyTypeStore = new Set());
        allIdents.set(dependencyIdent.identHash, dependencyIdent);
        byDependencyTypeStore.add(dependencyRange);
    }
    for (const [workspace, byWorkspacesStore] of byWorkspaces) {
        for (const [identHash, byIdentStore] of byWorkspacesStore) {
            const dependencyIdent = allIdents.get(identHash);
            if (typeof dependencyIdent === `undefined`)
                throw new Error(`Assertion failed: The ident should have been registered`);
            for (const [dependencyType, byDependencyTypeStore] of byIdentStore) {
                const expectedRanges = [...byDependencyTypeStore];
                if (expectedRanges.length > 2) {
                    report.reportError(core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} via conflicting ranges ${expectedRanges.slice(0, -1).map(expectedRange => core_3.structUtils.prettyRange(configuration, String(expectedRange))).join(`, `)}, and ${core_3.structUtils.prettyRange(configuration, String(expectedRanges[expectedRanges.length - 1]))} (in ${dependencyType})`);
                }
                else if (expectedRanges.length > 1) {
                    report.reportError(core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} via conflicting ranges ${core_3.structUtils.prettyRange(configuration, String(expectedRanges[0]))} and ${core_3.structUtils.prettyRange(configuration, String(expectedRanges[1]))} (in ${dependencyType})`);
                }
                else {
                    const dependencyDescriptor = workspace.manifest[dependencyType].get(dependencyIdent.identHash);
                    const [expectedRange] = expectedRanges;
                    if (expectedRange !== null) {
                        if (!dependencyDescriptor) {
                            if (fix) {
                                workspace.manifest[dependencyType].set(dependencyIdent.identHash, core_3.structUtils.makeDescriptor(dependencyIdent, expectedRange));
                                toSave.add(workspace);
                                hasFixes = true;
                            }
                            else {
                                report.reportError(core_2.MessageName.CONSTRAINTS_MISSING_DEPENDENCY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} (via ${core_3.structUtils.prettyRange(configuration, expectedRange)}), but doesn't (in ${dependencyType})`);
                            }
                        }
                        else if (dependencyDescriptor.range !== expectedRange) {
                            if (fix) {
                                workspace.manifest[dependencyType].set(dependencyIdent.identHash, core_3.structUtils.makeDescriptor(dependencyIdent, expectedRange));
                                toSave.add(workspace);
                                hasFixes = true;
                            }
                            else {
                                report.reportError(core_2.MessageName.CONSTRAINTS_INCOMPATIBLE_DEPENDENCY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} via ${core_3.structUtils.prettyRange(configuration, expectedRange)}, but uses ${core_3.structUtils.prettyRange(configuration, dependencyDescriptor.range)} instead (in ${dependencyType})`);
                            }
                        }
                    }
                    else {
                        if (dependencyDescriptor) {
                            if (fix) {
                                workspace.manifest[dependencyType].delete(dependencyIdent.identHash);
                                toSave.add(workspace);
                                hasFixes = true;
                            }
                            else {
                                report.reportError(core_2.MessageName.CONSTRAINTS_EXTRANEOUS_DEPENDENCY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} has an extraneous dependency on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} (in ${dependencyType})`);
                            }
                        }
                    }
                }
            }
        }
    }
    return hasFixes;
}
async function processFieldConstraints(toSave, enforcedFields, { configuration, fix, report }) {
    let hasFixes = false;
    const byWorkspaces = new Map();
    for (const { workspace, fieldPath, fieldValue } of enforcedFields) {
        const byWorkspacesStore = core_1.miscUtils.getMapWithDefault(byWorkspaces, workspace);
        const byPathStore = core_1.miscUtils.getSetWithDefault(byWorkspacesStore, fieldPath);
        byPathStore.add(fieldValue);
    }
    for (const [workspace, byWorkspacesStore] of byWorkspaces) {
        for (const [fieldPath, byPathStore] of byWorkspacesStore) {
            const expectedValues = [...byPathStore];
            if (expectedValues.length > 2) {
                report.reportError(core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to conflicting values ${expectedValues.slice(0, -1).map(expectedValue => configuration.format(String(expectedValue), `magenta`)).join(`, `)}, or ${configuration.format(String(expectedValues[expectedValues.length - 1]), `magenta`)}`);
            }
            else if (expectedValues.length > 1) {
                report.reportError(core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to conflicting values ${configuration.format(String(expectedValues[0]), `magenta`)} or ${configuration.format(String(expectedValues[1]), `magenta`)}`);
            }
            else {
                const actualValue = get_1.default(workspace.manifest.raw, fieldPath);
                const [expectedValue] = expectedValues;
                if (expectedValue !== null) {
                    if (actualValue === undefined) {
                        if (fix) {
                            await setWorkspaceField(workspace, fieldPath, expectedValue);
                            toSave.add(workspace);
                            hasFixes = true;
                        }
                        else {
                            report.reportError(core_2.MessageName.CONSTRAINTS_MISSING_FIELD, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to ${configuration.format(String(expectedValue), `magenta`)}, but doesn't`);
                        }
                    }
                    else if (JSON.stringify(actualValue) !== expectedValue) {
                        if (fix) {
                            await setWorkspaceField(workspace, fieldPath, expectedValue);
                            toSave.add(workspace);
                            hasFixes = true;
                        }
                        else {
                            report.reportError(core_2.MessageName.CONSTRAINTS_INCOMPATIBLE_FIELD, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${configuration.format(fieldPath, `cyan`)} set to ${configuration.format(String(expectedValue), `magenta`)}, but is set to ${configuration.format(JSON.stringify(actualValue), `magenta`)} instead`);
                        }
                    }
                }
                else {
                    if (actualValue !== undefined && actualValue !== null) {
                        if (fix) {
                            await setWorkspaceField(workspace, fieldPath, null);
                            toSave.add(workspace);
                            hasFixes = true;
                        }
                        else {
                            report.reportError(core_2.MessageName.CONSTRAINTS_EXTRANEOUS_FIELD, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} has an extraneous field ${configuration.format(fieldPath, `cyan`)} set to ${configuration.format(String(expectedValue), `magenta`)}`);
                        }
                    }
                }
            }
        }
    }
    return hasFixes;
}
async function setWorkspaceField(workspace, fieldPath, value) {
    if (value === null) {
        unset_1.default(workspace.manifest.raw, fieldPath);
    }
    else {
        set_1.default(workspace.manifest.raw, fieldPath, JSON.parse(value));
    }
}
